/*-------------------------------------------------------------------
 File description :
 cpp file for Gps_func class;
 Calculate sat visibility using Almanac file, current time, user's rough location
 Read GPS nav message, get ephemeris data;
 Calculate pseudorange;
 Resolve P.V.T. ( position, velocity, time bias );

--------------------------------------------------------------------*/

#include "Gpsfunc.h"
#include "CMatrix3.h"
#include <time.h>
#include <math.h>
#include <assert.h>
#include <string.h>


/*******************************************************************************
FUNCTION exor(char bit, long parity)
RETURNS  None.

PARAMETERS
			bit     char
			parity  long

PURPOSE
			count the number of bits set in the parameter parity and
			do an exclusive or with the parameter bit

WRITTEN BY
	Clifford Kelley
CHANGED BY
        Yu Lu

*******************************************************************************/
static int exor(char bit, unsigned int parity)
{
  char i;
  int result, tmp_par;
  result=0;
  tmp_par = parity>>6;
  for (i=0;i<24;i++)
  {
    if( tmp_par & 0x1) result++;
    tmp_par = tmp_par>>1;
  }
  result=result%2;

  result=(bit ^ result) & 0x1;
  return(result);
}

//------------------------------------------------------------------
// Class constructor and destructor

GPS_Func::GPS_Func()
{
 double pos[3];

 thrdtermed = true;
 for( int i=0; i<SatMax+1; i++)
 {
  Almanac[i].SatId = 0;  // id == 0 means this is not a valid data
  Ephemeris[i].valid = 0; // init epe validity to false
  eph_valid[i] = false;
  Ephemeris[i].iodc = -1; // invalid iodc
  Ephemeris[i].iode = -1; // invalid iode
 }

 for( int i=0; i<CH_NUM; i++)
  for( int j=0; j<5; j++)
  navmsgparity[i].par[j] = 0x1;  // init parity values for 5 frm of all channels

 memset(ch_prn, 0, CH_NUM*sizeof( unsigned char ));

 // default userposition is in Riverside, CA  :=)
 pos[0] = -117.3;  // longi
 pos[1] = 33.4;  // lati
 pos[2] = 500.;  // altitude

 // init user's position to invalid value;

 ECEFuserpos[0] = 1.;
 ECEFuserpos[1] = 1.;
 ECEFuserpos[2] = 1.;
 NEDuserpos[0] = 0.;      //long
 NEDuserpos[1] = 0.;      //lat
 NEDuserpos[2] = 0.;      //alt
 ECEFvel[0] = 0.;
 ECEFvel[1] = 0.;
 ECEFvel[2] = 0.;
 clk_bias = 0.;
 pdop=tdop=hdop=vdop=gdop=0.;

 // init user's accurate time to invalid value;
 AccurateTime = -1.;

 for( int i=0; i<CH_NUM; i++)
 {
  allch_navresult[i].prn = 0;
  allch_navresult[i].valid = false;
  allch_navresult[i].pseudorange = 0.;
  allch_navresult[i].ele = 0.;
  allch_navresult[i].azi = 0.;
 }

 SetUserPos(pos);
#ifdef GPS_FUNC_DBG
 debugfile.open(".\\log\\gps_func.log",ios::out);
#endif
}
GPS_Func::~GPS_Func()
{
#ifdef GPS_FUNC_DBG
 debugfile.close();
#endif
}
//------------------------------------------------------------------
// set longitude, latitude, altitude
// and calculate tangent plane parameters
void GPS_Func::SetUserPos( double* Pos)
{
 double SinLat, CosLat, SinLong, CosLong;
 UserPos[0] = Pos[0];  // longi
 UserPos[1] = Pos[1];  // lati
 UserPos[2] = Pos[2];  // altitude


 SinLat  = sin(Pos[1]*PI/180.);
 CosLat  = cos(Pos[1]*PI/180.);
 SinLong = sin(Pos[0]*PI/180.);
 CosLong = cos(Pos[0]*PI/180.);

 PlaneCE  =  CosLat;
 PlaneET1 = -SinLong;
 PlaneET2 =  CosLong;
 PlaneNT1 = -SinLat * CosLong;
 PlaneNT2 = -SinLat * SinLong;
 PlaneUTx =  CosLat * CosLong;
 PlaneUTy =  CosLat * SinLong;
 PlaneUTz =  SinLat;
 
}
//--------------------------------------------------------------------
//  PosECEF : destination position , in ECEF format
//  Pos:      souce position , in llh format
void GPS_Func::ConvertToECEF( double *PosECEF, double *Pos)
{
  double Dx = 0.0;
  double Dy = 0.0;
  double Dz = 0.0;

  double NormRadius, ActHeight;
  double CosLat, SinLat, CosLong, SinLong;
                 
  CosLong = cos( Pos[0] * PI / 180.0 );
  SinLong = sin( Pos[0] * PI / 180.0 );
  CosLat  = cos( Pos[1] * PI / 180.0 );
  SinLat  = sin( Pos[1] * PI / 180.0 );

  NormRadius = SemiMajorAxis /
    sqrt( 1.0 - ( EccentrSquared * SinLat * SinLat ) );
  ActHeight = NormRadius + Pos[2];

  PosECEF[0] = ActHeight * CosLat * CosLong - Dx;
  PosECEF[1] = ActHeight * CosLat * SinLong - Dy;
  PosECEF[2] = ( OneMinusEccentrSquared * NormRadius + Pos[2]) * SinLat - Dz;
}

//--------------------------------------------------------------------
// Pos:   Desitination position , in llh format
// PosECEF: Source position, in ECEF format

void GPS_Func::ConvertToLongLatAlt( double *Pos, double *PosECEF)
{
  double P, Q, R, D, SinL, SinL2;
  double X, Y, Z;
  double Long, Lat, Alt;

  X = PosECEF[0];
  Y = PosECEF[1];
  Z = PosECEF[2];

  Long = atan2( Y, X );

  P = sqrt( (X * X) + (Y * Y) );
  Q = Z / P;
  Alt = 0.0;

  Lat = atan2( Q, OneMinusEccentrSquared );

  do {
    SinL = sin(Lat);
    SinL2 = SinL * SinL;
    R = SemiMajorAxis / sqrt(1.0 - (EccentrSquared * SinL2));
    D = Alt;
    Alt = (P / cos(Lat)) - R;
    Lat = atan2(Q * (R + Alt), (OneMinusEccentrSquared * R) + Alt);
  }  while ( fabs(Alt - D) > 0.2 );

  Pos[0] = Long * 180.0 / PI;
  Pos[1] = Lat  * 180.0 / PI;
  Pos[2] = Alt;

}
//--------------------------------------------------------------------
//  sec : in 0-59;
//  min : in 0-59;
//  hr  : in 0-23;
//  day : in 1-31;
//  mon : in 1-12;
// calculate Julian day using sec,min,hr,day,mon,yr
long double GPS_Func::CalcJulianDay(int sec,int min,int hr,int day,int mon,int year)
{
 int e_year,e_mon;
 long double jd,h_jd, tmp_d;


 if( mon <=2){
  e_mon = mon +12;
  e_year = year -1;
 }
 else
 {
  e_mon = mon;
  e_year = year;
 }
 h_jd = (hr + min/60. + sec/3600.)/24.;
 tmp_d = (long double)(floor(365.25*e_year)) + (long double)(floor(30.6001*(e_mon+1)))+ (long double)day;
 jd =  tmp_d +  h_jd + 1720981.5L;
 return jd;
}
// calculate julian day using tm variable
long double GPS_Func::CalcJulianDay(struct tm * gt)
{
 int sec,min,hr,day,mon,year;

 sec = gt->tm_sec;
 min = gt->tm_min;
 hr  = gt->tm_hour;
 day = gt->tm_mday;
 mon = gt->tm_mon+1;
 year = gt->tm_year+1900;

 return CalcJulianDay(sec,min,hr,day,mon,year);
}
//--------------------------------------------------------------------
// calculate GPS week using Julian day
int GPS_Func::CalcGpsWeek(long double jd )
{
  int gw;
  gw = floor((jd-2444244.5)/7.); // 2444244.5 is JD of Jan,6,1980
  return (gw%1024);
}
// use current system time to calc gps time
void GPS_Func::get_gpsweek(void)
{
 struct tm *gmt;
 time_t curtime;
 double jd;

 time(&curtime);
 gmt = gmtime(&curtime);
 jd = CalcJulianDay(gmt);
 GpsWeek = CalcGpsWeek(jd);
}
// calculate time(in sec) of week using julian day
int GPS_Func::CalcToW( long double jd)
{
 int gw ;
 gw = floor((jd-2444244.5L)/7.);
 return( jd -2444244.5L - gw*7.)*SecPerDay ;
}
//--------------------------------------------------------------------
// read Almanac data for one satellite

void GPS_Func::read_almparam(AlmInfo *almpt, FILE *almfile)
{
  char info[1024],*begin_info;
  int tmp_int;
  double tmp_double;

  while( (fgets(info, 1024, almfile)) != NULL){
    if( (begin_info=strstr(info, ALM_HEALTH_STR)) ) break;
  }
  // we find health string 
  sscanf(begin_info, "Health:%d", &tmp_int);
  almpt->Health = tmp_int;
  
  while( (fgets(info, 1024, almfile)) != NULL){
    if( (begin_info=strstr(info, ALM_ECCEN_STR)) ) break;
  }
  // we find eccentricity str 
  sscanf(begin_info, "Eccentricity: %lg", &tmp_double);
  almpt->E = tmp_double;


  while( (fgets(info, 1024, almfile)) != NULL){
    if( (begin_info=strstr(info, ALM_TIME_STR)) ) break;
  }
  // we find time of availablity str 
  sscanf(begin_info, "Time of Applicability(s): %lg", &tmp_double);
  almpt->Toa = tmp_double;

  while( (fgets(info, 1024, almfile)) != NULL){
    if( (begin_info=strstr(info, ALM_ORB_STR)) ) break;
  }
  // we find ORB str
  sscanf(begin_info, "Orbital Inclination(rad): %le", &tmp_double);
  almpt->DeltaI = tmp_double;

  while( (fgets(info, 1024, almfile)) != NULL){
    if( (begin_info=strstr(info, ALM_RRA_STR)) ) break;
  }
  // we find RRA str 
  sscanf(begin_info, "Rate of Right Ascen(r/s): %le", &tmp_double);
  almpt->OmegaDot = tmp_double;

  while( (fgets(info, 1024, almfile)) != NULL){
    if( (begin_info=strstr(info, ALM_SQRTA_STR)) ) break;
  }
  // we find SQRTA str 
  sscanf(begin_info, "SQRT(A)  (m 1/2): %le", &tmp_double);
  almpt->SqrtA = tmp_double;

  while( (fgets(info, 1024, almfile)) != NULL){
    if( (begin_info=strstr(info, ALM_RAAW_STR)) ) break;
  }
  // we find RAAW str 
  sscanf(begin_info, "Right Ascen at Week(rad): %le", &tmp_double);
  almpt->Omega0 = tmp_double;

  while( (fgets(info, 1024, almfile)) != NULL){
    if( (begin_info=strstr(info, ALM_AOP_STR)) ) break;
  }
  // we find AOP str 
  sscanf(begin_info, "Argument of Perigee(rad): %le", &tmp_double);
  almpt->Omega = tmp_double;

  while( (fgets(info, 1024, almfile)) != NULL){
    if( (begin_info=strstr(info, ALM_MA_STR)) ) break;
  }
  // we find MA str 
  sscanf(begin_info, "Mean Anom(rad): %le",  &tmp_double);
  almpt->M0 = tmp_double;

  while( (fgets(info, 1024, almfile)) != NULL){
    if( (begin_info=strstr(info, ALM_AF0_STR)) ) break;
  }
  // we find AF0 str 
  sscanf(begin_info, "Af0(s): %le", &tmp_double);
  almpt->Af0 = tmp_double;

  while( (fgets(info, 1024, almfile)) != NULL){
    if( (begin_info=strstr(info, ALM_AF1_STR)) ) break;
  }
  // we find AF1 str 
  sscanf(begin_info, "Af1(s/s): %le", &tmp_double);
  almpt->Af1 = tmp_double;

  while( (fgets(info, 1024, almfile)) != NULL){
    if( (begin_info=strstr(info, ALM_WK_STR)) ) break;
  }
  // we find Week str 
  sscanf(begin_info, "week: %d", &tmp_int);
  almpt->WeekNo = tmp_int;
}
//--------------------------------------------------------------------
// read Almanac data for all satellites
void GPS_Func::ReadAlm(FILE *almfile)
{
  char info[1024],*begin_info, dummy[1024];
  int satid;

  AlmInfo  *alm;

  while( (fgets(info,1024,almfile)) != NULL){
    begin_info = strstr(info, ALM_ID_STR);  // find the beginning of ID str
    if( begin_info ){
      sscanf(begin_info, "%s %d", dummy, &satid); // read the id
      assert(satid <32);
      assert(satid >=0);
      alm = &Almanac[satid];
      alm->SatId = satid;
      read_almparam(alm, almfile);
    } // end of if( begin_info )
  }// end of while
}
//--------------------------------------------------------------------
// calculate sat position using ephemeris data and time of week
void GPS_Func::SatPosECEFEphinfo(EphInfo *EphPtr, double Time, int WeekNo, double *Pos)
{
 double d_toc,bclk,tc, d_toe, ei, ea, diff, aol, ta;
 double delr, delal, delinc, r, inc, la, xp, yp;

 d_toc = Time - EphPtr->toc;
 if(d_toc>302400.0)
  d_toc = d_toc - 604800.0;
 else if(d_toc<-302400.0)
  d_toc = d_toc + 604800.0;

 bclk = EphPtr->af0 + EphPtr->af1*d_toc + EphPtr->af2*d_toc*d_toc
	-EphPtr->tgd;
 tc = Time - bclk;

 d_toe = tc - EphPtr->toe;
 if(d_toe>302400.0)
  d_toe = d_toe-604800.0;
 else if(d_toe<-302400.0)
  d_toe = d_toe+604800.0;

 ei = EphPtr->ma + d_toe*(EphPtr->wm + EphPtr->dn);
 ea = ei;
 do
 {
   diff = (ei - (ea - EphPtr->ety*sin(ea)))/(1.0E0 - EphPtr->ety*cos(ea));
   ea = ea + diff;
 } while (fabs(diff) > 1.0e-12 );

 bclk = bclk - 4.442807633E-10*EphPtr->ety*EphPtr->sqra*sin(ea);
 EphPtr->bclk = bclk;

//     ea is the eccentric anomaly
 ta = atan2(sqrt(1.00-pow(EphPtr->ety,2))*sin(ea),cos(ea)-EphPtr->ety);

//     TA IS THE TRUE ANOMALY (ANGLE FROM PERIGEE)

 aol = ta + EphPtr->w;

//     AOL IS THE ARGUMENT OF LATITUDE OF THE SATELLITE

//     calculate the second harmonic perturbations of the orbit

 delr  = EphPtr->crc*cos(2.0*aol) + EphPtr->crs*sin(2.0*aol);
 delal = EphPtr->cuc*cos(2.0*aol) + EphPtr->cus*sin(2.0*aol);
 delinc= EphPtr->cic*cos(2.0*aol) + EphPtr->cis*sin(2.0*aol);

 //     R IS THE RADIUS OF SATELLITE ORBIT AT TIME T

 r = pow(EphPtr->sqra,2)*(1.00 - EphPtr->ety*cos(ea)) + delr;
 aol = aol + delal;
 inc = EphPtr->inc0 + delinc + EphPtr->idot*d_toe;

 //     LA IS THE CORRECTED LONGITUDE OF THE ASCENDING NODE

 la = EphPtr->w0 + (EphPtr->omegadot - OmegaDotEarth)*d_toe
       - OmegaDotEarth*EphPtr->toe;

 xp = r*cos(aol);
 yp = r*sin(aol);

 Pos[0] = xp*cos(la) - yp*cos(inc)*sin(la);
 Pos[1] = xp*sin(la) + yp*cos(inc)*cos(la);
 Pos[2] = yp*sin(inc);
}

//--------------------------------------------------------------------
// calculate sat position using almanac data and time of week
void GPS_Func::SatPosECEFAlminfo(AlmInfo *AlmPtr, double Time, int WeekNo, double *Pos)
{
  double d_toa, ei,ea,diff, m_m, ta ,r, aol, la, omegae,xp,yp;
  double satPos[3];

//  omegae = 7.2921151467E-5;
  d_toa = Time - AlmPtr->Toa;
  if( d_toa > 302400.0)
    d_toa = d_toa - 604800.0;
  else if( d_toa < -302400.0)
    d_toa = d_toa + 604800.0;
  // mean motion is
  m_m = 19964981.84/pow(AlmPtr->SqrtA,3);
  // recursively calc eccentric anomaly E

  ei = AlmPtr->M0 + d_toa*m_m;
  ea = ei;
  do{
    diff = (ei - (ea - AlmPtr->E*sin(ea)))/(1. - AlmPtr->E*cos(ea));
    ea = ea + diff;
  } while( fabs(diff) > 1.0e-6);

  // Now ea is eccentric anomaly E

  //ta is the true anomaly
  if( AlmPtr->E != 0.0)
    ta = atan2(sqrt(1. - AlmPtr->E * AlmPtr->E) * sin(ea),cos(ea)-AlmPtr->E);
  else
    ta = ea;

  //r is the radius of satellite orbit at time Time
  r = AlmPtr->SqrtA*AlmPtr->SqrtA*(1.0 - AlmPtr->E*cos(ea));

  // aol is the argument of latitude
  aol = ta + AlmPtr->Omega;

  //
  la = AlmPtr->Omega0 + (AlmPtr->OmegaDot-OmegaDotEarth)*d_toa
       - AlmPtr->Toa*OmegaDotEarth;
  xp= r*cos(aol);
  yp= r*sin(aol);

  Pos[0] = xp*cos(la) - yp*cos(AlmPtr->DeltaI)*sin(la);
  Pos[1] = xp*sin(la) + yp*cos(AlmPtr->DeltaI)*cos(la);
  Pos[2] = yp*sin(AlmPtr->DeltaI);

  AlmPtr->ECEF_Pos[0] = Pos[0];
  AlmPtr->ECEF_Pos[1] = Pos[1];
  AlmPtr->ECEF_Pos[2] = Pos[2];


  ConvertToLongLatAlt(  satPos,  Pos);
}

/*++
 Routine Description:
   calculate sat's position using ephemeris and almanac for comparition;
   just for debugging use;
 Argument:
         prn: prn number of satellite;
         Time: time of gps
 Rreturn:
         none;

--*/
void GPS_Func::GetSatPos( int prn , double Time)
{
 double posalm[3], poseph[3];
 SatPosECEFAlminfo(&Almanac[prn], Time, 0, posalm);
 SatPosECEFEphinfo(&Ephemeris[prn], Time, 0, poseph);

//#ifdef GPS_FUNC_DBG
//   debugfile << "TOW: " << Time << endl;
//   debugfile.setf(10);
//   debugfile << "Pos from Alm :"  << posalm[0] << ","
//                                   << posalm[1] << ","
//                                   << posalm[2] << endl
//             <<  "Pos from Eph :"  << poseph[0] << ","
//                                   << poseph[1] << ","
//                                   << poseph[2] << endl;
//
//#endif
}
//--------------------------------------------------------------------

void GPS_Func::SatElevAndAzimuth( int SatId, double UsrPos[3],
                                  double Time,int Week )
{
  double Dx, Dy, Dz;         /* Distance user to sat in ECEF */
  double SatPos[3],SatPosPlus[3],SatSpeed[3];  /* Position of sat */
  double UserToSatRange;     /* Distance user to sat */
  double DirCos[3];          /* Director cosines from user to sat */
  double EnuCos[3];

  /* Calculate the sat pos from almanac, current time and week number */
  SatPosECEFAlminfo( &Almanac[SatId], Time, Week, SatPos );
  SatPosECEFAlminfo( &Almanac[SatId], Time+1., Week, SatPosPlus );
  for(int k =0; k<3; k++)
  SatSpeed[k] = SatPosPlus[k] - SatPos[k];

  /* Distance from user to sat */
  Dx = SatPos[0] - UsrPos[0];
  Dy = SatPos[1] - UsrPos[1];
  Dz = SatPos[2] - UsrPos[2];

  UserToSatRange = sqrt((Dx * Dx) + (Dy * Dy) + (Dz * Dz));

  Almanac[SatId].Dopp= (SatSpeed[0]*Dx + SatSpeed[1]*Dy+SatSpeed[2]*Dz)*5.2514/UserToSatRange;

  /* Director cosines */
  DirCos[0] = Dx / UserToSatRange;
  DirCos[1] = Dy / UserToSatRange;
  DirCos[2] = Dz / UserToSatRange;

  /* Transformation to local tangent plane */
  EnuCos[0] = (PlaneET1 * DirCos[0]) + (PlaneET2 * DirCos[1]);
  EnuCos[1] = (PlaneNT1 * DirCos[0]) + (PlaneNT2 * DirCos[1])
    + (PlaneCE  * DirCos[2]);
  EnuCos[2] = (PlaneUTx * DirCos[0]) + (PlaneUTy * DirCos[1])
    + (PlaneUTz * DirCos[2]);

  /* Azimuth and elevation of sat as seen by user */
  Almanac[SatId].Ele = asin(EnuCos[2]) * (180.0 / PI);
  Almanac[SatId].Azm = atan2(EnuCos[0], EnuCos[1]) * (180.0 / PI);
}
//--------------------------------------------------------------------
/*++
 Routine Description:
   user alamanac file to calculate sat's elevation and azimuth roughly;
 Argument:
         time: time in second ;
 Rreturn:
         none;

--*/
void GPS_Func::process_allsat(double Time, int Week)
{
 double MyECEFPos[3];
 // first init ele, azm and doppler to invalid value
 for( int i=0; i< SatMax+1; i++)
 {
  Almanac[i].Ele = -1;
  Almanac[i].Azm = -1;
  Almanac[i].Dopp = 0;
 }

 ConvertToECEF(MyECEFPos, UserPos );

 for( int i=0; i<SatMax+1; i++)
 {
  if( Almanac[i].SatId != 0)
  { // this is a valid almanac data
   SatElevAndAzimuth(i,MyECEFPos,Time,Week);
  }
 }
}

//--------------------------------------------------------------------
/*++
 Routine Description:
   process GPS message, using global variable allch_nagmsg to transfer msgs
 Argument:
         none;
 Rreturn:
         none;

--*/
void GPS_Func::processMsg( void )
{
 int ch_cnt;

 if( newmsg_rdy )
 {
  for(int i=0; i<CH_NUM; i++)
   if( allch_navmsg[i].newdata ) // new frame is here
   {
    ch_prn[i] = allch_navmsg[i].prn; // set prn num for this channel
    readOneChMsg( i );
    allch_navmsg[i].newdata = false; // reset new flag
   }
  newmsg_rdy = false;

 }

 ch_cnt = allchtrtime.ch_count;

//#ifdef GPS_FUNC_DBG
// debugfile << " CH count is :" <<  ch_cnt << endl;
// debugfile.precision(16);
// for( int i=0; i<ch_cnt; i++)
//  debugfile << allchtrtime.ch_transtime[i].prn << " "
//            << allchtrtime.ch_transtime[i].tr_time << endl;
//#endif

 if( ch_cnt >= 4)
 {
  resolve_PVT();
 }
}

//--------------------------------------------------------------------
/*++
 Routine Description:
      P.V.T. resolution;
 Argument:
      none;
 Rreturn:
         none;

--*/

void GPS_Func::resolve_PVT(void)
{
 long double min_trtime, max_trtime,m_time;
 int prn;
 sat_tr_pos sat_pvtpara[CH_NUM];
 double tmp_pos[3];

 // first check if all trans_time make sense?
 min_trtime = max_trtime = allchtrtime.ch_transtime[0].tr_time;

 for( int i=1; i< allchtrtime.ch_count; i++)
 {
  if( allchtrtime.ch_transtime[i].tr_time > max_trtime)
   max_trtime = allchtrtime.ch_transtime[i].tr_time;
  if( allchtrtime.ch_transtime[i].tr_time < min_trtime)
   min_trtime = allchtrtime.ch_transtime[i].tr_time;
 }

 if( fabs( max_trtime - min_trtime ) > 0.1 )    // doesn't make sense
  return;

 // if any channel is not ready , quit
 for( int i=0; i< allchtrtime.ch_count; i++)
 {
  prn = allchtrtime.ch_transtime[i].prn;
  if( ! Ephemeris[prn].valid)
   return;
 }

 m_time = (long double)( max_trtime + 0.08);  // guess

 for( int i=0; i< allchtrtime.ch_count; i++)
 {
  prn = allchtrtime.ch_transtime[i].prn;
  SatPosECEFEphinfo(&Ephemeris[prn],
                     allchtrtime.ch_transtime[i].tr_time,
                     0,
                     &sat_pvtpara[i].pos[0]);

  SatPosECEFEphinfo(&Ephemeris[prn],
                     allchtrtime.ch_transtime[i].tr_time - 1.,
                     0,
                     &tmp_pos[0]);
                     
  // calculate velocity of satellite

   sat_pvtpara[i].vel[0] = sat_pvtpara[i].pos[0] - tmp_pos[0]-
                           sat_pvtpara[i].pos[1]* OmegaDotEarth;
   sat_pvtpara[i].vel[1] = sat_pvtpara[i].pos[1] - tmp_pos[1]+
                           sat_pvtpara[i].pos[0]* OmegaDotEarth;
   sat_pvtpara[i].vel[2] = sat_pvtpara[i].pos[2] - tmp_pos[2];


  sat_pvtpara[i].delta_time = m_time - allchtrtime.ch_transtime[i].tr_time + Ephemeris[prn].bclk;
  sat_pvtpara[i].doppler_freq = allchtrtime.ch_transtime[i].dopp_f;

#ifdef GPS_FUNC_DBG
 debugfile.precision(16);
 debugfile << "Prn:" <<prn << " "
            << " x: " << sat_pvtpara[i].pos[0]
            << " y: " << sat_pvtpara[i].pos[1]
            << " z: " << sat_pvtpara[i].pos[2]
            << " vx:" <<sat_pvtpara[i].vel[0]
            << " vy:" <<sat_pvtpara[i].vel[1]
            << " vz:" <<sat_pvtpara[i].vel[2]
            << " tr:" << allchtrtime.ch_transtime[i].tr_time << endl;
#endif

 }

 // next we will use recursion to get current location
 int nits, i, j, k, nsl;
 double t,x,y,z;
 double  dd[4][4],r,ms[4][6],pm[4][6],bm[6],br[4],correct_mag;
 double a1,b1,c1,d1,e1,f1,g1,h1,i1,j1,k1,l1,m1,n1,o1,p1,denom,alpha;

 double Dx, Dy, Dz;         /* Distance user to sat in ECEF */
 double UserToSatRange;     /* Distance user to sat */
 double DirCos[3];          /* Director cosines from user to sat */
 double EnuCos[3];


 nsl = allchtrtime.ch_count;
 nits=0;
 t=0.0;
 x=ECEFuserpos[0];
 y=ECEFuserpos[1];
 z=ECEFuserpos[2];
 do
 {
  for(i=0;i<nsl;i++)
  {

//   Compute range in ECI at the time of arrival at the receiver

   alpha = (t-sat_pvtpara[i].delta_time)*OmegaDotEarth;
   r = sqrt(pow(sat_pvtpara[i].pos[0]*cos(alpha)-sat_pvtpara[i].pos[1]*sin(alpha)-x,2)+
	    pow(sat_pvtpara[i].pos[1]*cos(alpha)+sat_pvtpara[i].pos[0]*sin(alpha)-y,2)+
	    pow(sat_pvtpara[i].pos[2]-z,2));
   bm[i]=r-(sat_pvtpara[i].delta_time - t)*SpeedOfLight;
   ms[0][i]=(sat_pvtpara[i].pos[0]*cos(alpha)-sat_pvtpara[i].pos[1]*sin(alpha)-x)/r;
   ms[1][i]=(sat_pvtpara[i].pos[1]*cos(alpha)+sat_pvtpara[i].pos[0]*sin(alpha)-y)/r;
   ms[2][i]=(sat_pvtpara[i].pos[2]-z)/r;
   ms[3][i]=1.0;
  }
  a1=0.;b1=0.;c1=0.;d1=0.;
  e1=0.;f1=0.;g1=0.;h1=0.;
  i1=0.;j1=0.;k1=0.;l1=0.;
  m1=0.;n1=0.;o1=0.;p1=0.;
  for (k=0;k<nsl;k++)
  {
   a1+=ms[0][k]*ms[0][k];
   b1+=ms[0][k]*ms[1][k];
   c1+=ms[0][k]*ms[2][k];
   d1+=ms[0][k]*ms[3][k];
//     e1+=ms[2][k]*ms[1][k];   for completeness, the matrix is symmetric
   f1+=ms[1][k]*ms[1][k];
   g1+=ms[1][k]*ms[2][k];
   h1+=ms[1][k]*ms[3][k];
//     i1+=ms[3][k]*ms[1][k];
//     j1+=ms[3][k]*ms[2][k];
   k1+=ms[2][k]*ms[2][k];
   l1+=ms[2][k]*ms[3][k];
//     m1+=ms[1][k];
//     n1+=ms[2][k];
//     o1+=ms[3][k];
   p1+=ms[3][k];
  }
  o1=l1;m1=d1;n1=h1;e1=b1;i1=c1;j1=g1;
/*
	  SOLVE FOR THE MATRIX INVERSE
*/
  denom=(k1*p1-l1*o1)*(a1*f1-b1*e1) + (l1*n1-j1*p1)*(a1*g1-c1*e1) +
        (j1*o1-k1*n1)*(a1*h1-d1*e1) + (l1*m1-i1*p1)*(c1*f1-b1*g1) +
        (i1*o1-k1*m1)*(d1*f1-b1*h1) + (i1*n1-j1*m1)*(c1*h1-d1*g1);

  dd[0][0]=f1*(k1*p1-l1*o1)+g1*(l1*n1-j1*p1)+h1*(j1*o1-k1*n1);
  dd[0][1]=e1*(l1*o1-k1*p1)+g1*(i1*p1-l1*m1)+h1*(k1*m1-i1*o1);
  dd[0][2]=e1*(j1*p1-n1*l1)-i1*(f1*p1-n1*h1)+m1*(f1*l1-j1*h1);
  dd[0][3]=e1*(n1*k1-j1*o1)+i1*(f1*o1-n1*g1)+m1*(j1*g1-f1*k1);

  dd[1][0]=dd[0][1];
  dd[1][1]=a1*(k1*p1-l1*o1)+c1*(l1*m1-i1*p1)+d1*(i1*o1-k1*m1);
  dd[1][2]=a1*(l1*n1-j1*p1)+i1*(b1*p1-n1*d1)+m1*(j1*d1-b1*l1);
  dd[1][3]=a1*(j1*o1-n1*k1)-i1*(b1*o1-n1*c1)+m1*(b1*k1-c1*j1);

  dd[2][0]=dd[0][2];
  dd[2][1]=dd[1][2];
  dd[2][2]=a1*(f1*p1-h1*n1)+b1*(h1*m1-e1*p1)+d1*(e1*n1-f1*m1);
  dd[2][3]=a1*(n1*g1-f1*o1)+e1*(b1*o1-c1*n1)+m1*(c1*f1-b1*g1);

  dd[3][0]=dd[0][3];
  dd[3][1]=dd[1][3];
  dd[3][2]=dd[2][3];
  dd[3][3]=a1*(f1*k1-g1*j1)+b1*(g1*i1-e1*k1)+c1*(e1*j1-f1*i1);

  if ( denom<=0.0 )
  {
   /*
    result.x=1.0;      // something went wrong
    result.y=1.0;      // set solution to near center of earth
    result.z=1.0;
    result.dt=0.0;
   */
  }
  else
  {
   for (i=0;i<4;i++)
   {
    for (j=0;j<4;j++)
     dd[i][j]=dd[i][j]/denom;
   }
   for (i=0;i<nsl;i++)
   {

    for (j=0;j<4;j++)
    {
     pm[j][i]=0.0;
     for (k=0;k<4;k++)
      pm[j][i]+=dd[j][k]*ms[k][i];
    }
   }
   for (i=0;i<4;i++)
   {
    br[i]=0.0;
    for (j=0;j<nsl;j++)
     br[i]+=bm[j]*pm[i][j];
   }
   nits++;
   x=x+br[0];
   y=y+br[1];
   z=z+br[2];
   t=t-br[3]/SpeedOfLight;
   correct_mag=sqrt(br[1]*br[1]+br[2]*br[2]+br[0]*br[0]);
  }
 } while ( correct_mag > 0.01 && correct_mag < 1.e8 && nits < 10);


// set position

 ECEFuserpos[0] = x;
 ECEFuserpos[1] = y;
 ECEFuserpos[2] = z;

 ConvertToLongLatAlt(NEDuserpos, ECEFuserpos);
 SetUserPos( NEDuserpos  );
 AccurateTime = m_time - t;

// for velocity
  for (i=0;i<nsl;i++)
  {
	 alpha=( AccurateTime - allchtrtime.ch_transtime[i].tr_time)*OmegaDotEarth;
	 r=sqrt(pow(sat_pvtpara[i].pos[0]*cos(alpha)-sat_pvtpara[i].pos[1]*sin(alpha)-x,2)+
			  pow(sat_pvtpara[i].pos[1]*cos(alpha)+sat_pvtpara[i].pos[0]*sin(alpha)-y,2)+
			  pow(sat_pvtpara[i].pos[2]-z,2));
	 bm[i]=((sat_pvtpara[i].pos[0]*cos(alpha)-sat_pvtpara[i].pos[1]*sin(alpha)-x)*sat_pvtpara[i].vel[0]+
			  (sat_pvtpara[i].pos[1]*cos(alpha)+sat_pvtpara[i].pos[0]*sin(alpha)-y)*sat_pvtpara[i].vel[1]+
			  (sat_pvtpara[i].pos[2]-z)*sat_pvtpara[i].vel[2])/r -
                           sat_pvtpara[i].doppler_freq*lambda;
  }

  for ( i=0;i<4;i++)
  {
	  br[i]=0.0;
	  for (j=0;j<nsl;j++)br[i]+=bm[j]*pm[i][j];
  }

 ECEFvel[0] = br[0] + y*OmegaDotEarth;
 ECEFvel[1] = br[1] - x*OmegaDotEarth;
 ECEFvel[2] = br[2];
 clk_bias = br[3]/SpeedOfLight*1000000.0;

// then calculate matrix G for dops
 Matrix H(nsl,4),G(4,4);

 EnterCriticalSection(&GpsCriticalSection);
  for( int i=0; i< allchtrtime.ch_count; i++)
 {
  allch_navresult[i].prn = allchtrtime.ch_transtime[i].prn;
  allch_navresult[i].valid = true;
  // calculate pseudorange
  allch_navresult[i].pseudorange =( AccurateTime -
                                 allchtrtime.ch_transtime[i].tr_time)*
                                 SpeedOfLight;
  // calculate elevation and azimuth


  /* Distance from user to sat */
  Dx = sat_pvtpara[i].pos[0] - UserPos[0];
  Dy = sat_pvtpara[i].pos[1] - UserPos[1];
  Dz = sat_pvtpara[i].pos[2] - UserPos[2];

  UserToSatRange = sqrt((Dx * Dx) + (Dy * Dy) + (Dz * Dz));

  H(i+1,1)= Dx/UserToSatRange;
  H(i+1,2)= Dy/UserToSatRange;
  H(i+1,3)= Dz/UserToSatRange;
  H(i+1,4)= 1.0;


   /* Director cosines */
  DirCos[0] = Dx / UserToSatRange;
  DirCos[1] = Dy / UserToSatRange;
  DirCos[2] = Dz / UserToSatRange;

  /* Transformation to local tangent plane */
  EnuCos[0] = (PlaneET1 * DirCos[0]) + (PlaneET2 * DirCos[1]);
  EnuCos[1] = (PlaneNT1 * DirCos[0]) + (PlaneNT2 * DirCos[1])
    + (PlaneCE  * DirCos[2]);
  EnuCos[2] = (PlaneUTx * DirCos[0]) + (PlaneUTy * DirCos[1])
    + (PlaneUTz * DirCos[2]);

  /* Azimuth and elevation of sat as seen by user */
  allch_navresult[i].ele = asin(EnuCos[2]) * (180.0 / PI);
  allch_navresult[i].azi = atan2(EnuCos[0], EnuCos[1]) * (180.0 / PI);

 }
 G=(H.t()*H).i();

 hdop=sqrt(G(1,1)+G(2,2));
 vdop=sqrt(G(3,3));
 tdop=sqrt(G(4,4));
 pdop=sqrt(G(1,1)+G(2,2)+G(3,3));
 gdop=sqrt(G(1,1)+G(2,2)+G(3,3)+G(4,4));

 LeaveCriticalSection(&GpsCriticalSection);



 #ifdef GPS_FUNC_DBG
 debugfile << " Pos of User:"
 << x << " " << y << " " << z <<" " << t<< " "
 <<nits << " " << correct_mag<<endl;
 #endif
}

//--------------------------------------------------------------------
/*++
 Routine Description:
   Process one channel's GPS message, idx is the channel's index
    using global variable allch_nagmsg to transfer msgs

 Argument:
         idx :  indicate which channel to read
 Rreturn:
         none;

--*/

void GPS_Func::readOneChMsg(int idx)
{
 int cur_idx;
 cur_idx = (allch_navmsg[idx].index + 1)%302;

 // pre_bit[0-1] are used to check paritys
 prev_bit[0] = (allch_navmsg[idx].msg[ cur_idx ])>>29;
 prev_bit[1] = (allch_navmsg[idx].msg[ (++cur_idx)%302 ])>>29;

 // re-organize the message bits to form message words
 for( int wi = 0; wi < 10; wi++ )
 {
  nav_msg[idx][wi] = 0;
  for( int j=0; j<30; j++)
  {
   if( allch_navmsg[idx].msg[(++cur_idx)%302] & 0x20000000 )
    nav_msg[idx][wi] = nav_msg[idx][wi]*2 + 0x1;
   else
    nav_msg[idx][wi] = nav_msg[idx][wi]*2 ;
  }
 }
  // till now , we have read all data for this channel in this time

 parity_check(idx);

 if( p_error[idx] == 0) // only when parity check passed
 {
  int frmidx = nav_msg[idx][1];
  frmidx = (frmidx>>2) & 0x07;
  if( frmidx >=1 && frmidx<=5 )
  {
//#ifdef GPS_FUNC_DBG
//   debugfile << hex << "CH #" << idx << ": ";
//   for( int k=0;k<10; k++)
//    debugfile <<  nav_msg[idx][k]<< " ";
//   debugfile << ". Parity: " << p_error[idx] << " frmidx:" << frmidx << endl;
//#endif
   navmsgparity[idx].par[frmidx-1] = 0;
   memcpy((void*)&navmsgparity[idx].msg[frmidx-1][0],(void*)&nav_msg[idx][0],10*sizeof(unsigned int));
   if( frmidx == 5) // we get all 5 subframes
    read_allsubframe(idx);
  }
 }
}

//--------------------------------------------------------------------
/*++
 Routine Description:
    Check parity for one subframe
     For details algorithmm, refer to ICD_GPS_200C
 Argument:
         ch_i :  indicate which channel to read
 Rreturn:
         none;

--*/
void GPS_Func::parity_check(int ch_i)
{
 char b_1, b_2, b_3, b_4, b_5, b_6;
 int parity, m_parity;
 p_error[ch_i] = 0x0;

 for( int word=0; word<10; word++)
 {
       m_parity = nav_msg[ch_i][word] & 0x3f;
       b_1=exor(prev_bit[0],nav_msg[ch_i][word] & pb1) *32;
       b_2=exor(prev_bit[1],nav_msg[ch_i][word] & pb2) *16;
       b_3=exor(prev_bit[0],nav_msg[ch_i][word] & pb3) *8;
       b_4=exor(prev_bit[1],nav_msg[ch_i][word] & pb4) *4;
       b_5=exor(0,nav_msg[ch_i][word] & pb5) *2;
       b_6=exor(prev_bit[0]^prev_bit[1],nav_msg[ch_i][word] & pb6);
       parity=b_1+b_2+b_3+b_4+b_5+b_6;
       if( parity != m_parity ) // parity check failed
        p_error[ch_i] = p_error[ch_i]*2 +1;
       else
        p_error[ch_i] = p_error[ch_i]*2;

       if( prev_bit[1] )
        nav_msg[ch_i][word] = nav_msg[ch_i][word]^0x3fffffc0;
       nav_msg[ch_i][word] = nav_msg[ch_i][word]>>6;
       prev_bit[0]=(m_parity & 0x2) >>1;
       prev_bit[1]= m_parity & 0x1;
 }

}
//--------------------------------------------------------------------
/*++
 Routine Description:
   Read nav msg for one channel,
    get ephemeris data for this channel.
 Argument:
         idx :  indicate which channel to read
 Rreturn:
         none;

--*/
void GPS_Func::read_allsubframe(int idx)
{
 int iodc,iode,idoe, iweek, iura, ihealth, iaf0, im0, inc0, iomega0,iw;
 int iomegadot, idot, prn;
 char itgd,iaf2;
 unsigned int itoc, ie,isqra, itoe;
 short int iaf1,icrs, idn, icuc,icus,icic,icis,icrc;
 unsigned int *sf;
 double r_sqra,r_inc0,r_ety;

 sf = (unsigned int*)&navmsgparity[idx].msg[0][0];
 prn = ch_prn[idx];

 if( navmsgparity[idx].par[0] == 0 &&
     navmsgparity[idx].par[1] == 0 &&
     navmsgparity[idx].par[2] == 0) // if frm1-3 is correct
     {
      iodc=int(((sf[0*10+2] & 0x3) <<8 ) | ((sf[0*10+7] & 0xFF0000) >>16));
      iode=int(sf[1*10+2]  >> 16);
      idoe=int(sf[2*10+9] >> 16);

      if( (iode == idoe) &&
          ( (iode!= Ephemeris[ prn ].iode) || (iodc!=Ephemeris[ prn ].iodc) ))
      {

       // subframe 1

       iweek= int(sf[0*10+2] >> 14);
       iura=int(( sf[0*10+2] & 0xF00 ) >> 8);
       ihealth=int(( sf[0*10+2] & 0xFC ) >> 2);
       itgd=sf[0*10+6] & 0xFF;
       itoc=sf[0*10+7] & 0xFFFF;
       iaf2=sf[0*10+8] >> 16;
       iaf1=sf[0*10+8] & 0xFFFF;
       iaf0=sf[0*10+9] >> 2;
       if( iaf0 & 0x200000) iaf0=iaf0 | 0xFFC00000; // sign extension

       //   subframe 2
       icrs=sf[1*10+2] & 0xFFFF;
       idn =(sf[1*10+3]>>8) & 0xFFFF;
       im0=((sf[1*10+3] & 0xFF) << 24) | sf[1*10+4];
       icuc=(sf[1*10+5]>>8) & 0xFFFF;
       ie=((sf[1*10+5] & 0xFF) << 24) | sf[1*10+6];
       icus=(sf[1*10+7]>>8) & 0xFFFF;
       isqra=(((sf[1*10+7] & 0xFF) << 24) | sf[1*10+8]);
       itoe=int(sf[1*10+9] >> 8);

       // subframe 3
       icic=(sf[2*10+2]>>8) & 0xFFFF;
       icis=(sf[2*10+4]>>8) & 0xFFFF;
       inc0=((sf[2*10+4] & 0xFF) << 24) | sf[2*10+5];
       iomega0=((sf[2*10+2] & 0xFF) << 24) | sf[2*10+3];
       icrc=(sf[2*10+6]>>8) & 0xFFFF;
       iw=((sf[2*10+6] & 0xFF) << 24) | sf[2*10+7];
       iomegadot=sf[2*10+8];
       if (iomegadot&0x800000)
        iomegadot=iomegadot | 0xFF000000;  // sign extension
       idot=((sf[2*10+9] & 0xFFFC) >> 2);
       if (idot&0x2000)
        idot=idot | 0xFFFFC000; // sign extension

       r_sqra=isqra*c_2m19;
       r_inc0=inc0*c_2m31*PI;
       r_ety=ie*c_2m33;

       if (   (r_inc0<1.05 && r_inc0>0.873)&&
              (r_sqra>5100.0 && r_sqra<5200.0)&&
              (r_ety <.05 && r_ety>0.0)  )
        {
          Ephemeris[prn].valid=1;
          eph_valid[prn] = true;  //set this global variable for gpsnav thread use

          Ephemeris[prn].iode=iode;
          Ephemeris[prn].iodc=iodc;
	  Ephemeris[prn].week=iweek;
	  Ephemeris[prn].ura=iura;
	  Ephemeris[prn].health=ihealth;
	  Ephemeris[prn].tgd=itgd*c_2m31;
	  Ephemeris[prn].toc=itoc*16.0;
	  Ephemeris[prn].af2=iaf2*c_2m55;
	  Ephemeris[prn].af1=iaf1*c_2m43;
	  Ephemeris[prn].af0=iaf0*c_2m31;
	  Ephemeris[prn].crs=icrs*c_2m5;
	  Ephemeris[prn].dn=idn*c_2m43*PI;
	  Ephemeris[prn].ma=im0*c_2m31*PI;
	  Ephemeris[prn].cuc=icuc*c_2m29;
	  Ephemeris[prn].ety=r_ety;
	  Ephemeris[prn].cus=icus*c_2m29;
	  Ephemeris[prn].sqra=r_sqra;
	  Ephemeris[prn].wm=19964981.84/pow(r_sqra,3);
	  Ephemeris[prn].toe=itoe*c_2p4;
	  Ephemeris[prn].cic=icic*c_2m29;
	  Ephemeris[prn].cis=icis*c_2m29;
	  Ephemeris[prn].inc0=r_inc0;
	  Ephemeris[prn].w0=iomega0*c_2m31*PI;
	  Ephemeris[prn].crc=icrc*c_2m5;
	  Ephemeris[prn].w=iw*c_2m31*PI;
	  Ephemeris[prn].omegadot=iomegadot*c_2m43*PI;
	  Ephemeris[prn].idot=idot*c_2m43*PI;
        } // end of  if (   (r_inc0<1.05 && r_inc0>0.873)&&
      } // end of if ( (iode == idoe) &&

     } // end of if( navmsgparity[idx].par[0] == 0 &&


  // finally init parity to invalid values back
  for( int j=0; j<5; j++)
  {
   navmsgparity[idx].par[j] = 0x1;  // init parity values for 5 frm of all channels
   memset((void*)&navmsgparity[idx].msg[j][0],0,10*sizeof(unsigned int));
  }
}
//--------------------------------------------------------------------
/*++
 Routine Description:
   Get result of processing ,
    such as user's pos, user's velocity, accurate time, etc.
 Argument:
         nav_pt:  pointer for processing result storing;
 Rreturn:
         none;

--*/

void GPS_Func::get_navinfo( PT_navinfo_disp nav_pt)
{
 memcpy(nav_pt->ECEFpos, ECEFuserpos, 3*sizeof(double));
 memcpy(nav_pt->NEDpos, NEDuserpos, 3*sizeof(double));
 memcpy(nav_pt->ECEFvel, ECEFvel, 3*sizeof(double));
 nav_pt->clk_bias = clk_bias;
 nav_pt->rcv_time = AccurateTime;
 nav_pt->GDOP = gdop;
 nav_pt->TDOP = tdop;
 nav_pt->PDOP = pdop;
 nav_pt->HDOP = hdop;
 nav_pt->VDOP = vdop;
}

//--------------------------------------------------------------------
/*++
 Routine Description:
  Set prn allocation according to sat's elevation.
 Argument:
         pn_num:  prn allocation buffer pointer;
 Rreturn:
         none;
--*/

void GPS_Func::SelectPrn(int* pn_num)
{
 int prn_idx[SatMax+1],tmp_id, sat_count, flag ;
 double ele[SatMax+1], tmp_ele;

 sat_count = 0;
 memset(prn_idx, 0, sizeof(prn_idx));

 for(int i=0 ; i<  SatMax+1; i++)
 {
  if( Almanac[i].SatId != 0 && Almanac[i].Ele > 0.) // a valid satellite
  {
   prn_idx[sat_count] = Almanac[i].SatId;
   ele[sat_count] = Almanac[i].Ele;
   sat_count++;
  }
 }
 // bubble sort to find the 6 biggest elevations
 do{
  flag = 0;
  for( int i=1; i< sat_count; i++)
   if( ele[i] > ele[i-1] )
   {   // swap i and i-1
    tmp_id = prn_idx[i];
    tmp_ele = ele[i];

    prn_idx[i] = prn_idx[i-1];
    ele[i] = ele[i-1];

    prn_idx[i-1] = tmp_id;
    ele[i-1] = tmp_ele;

    flag = 1; // set flat to indicate we need further addjustment
   }
 }while( flag );

 for( int idx = 0; idx < 6 ; idx++)
  pn_num[idx] = prn_idx[idx];

}

//--------------------------------------------------------------------
/*++
 Routine Description:
  calculate position of sats through almanac data
   and set prn allocation accordingly.
 Argument:
         pn:  prn allocation buffer pointer;
         gt:  struct tm* pointer, pointing to time of setting;
         AlmFile: FILE pointer of almanac file;
         Pos: roughly guessed user position;
 Rreturn:
         0;

--*/

int GPS_Func::CalcPrnSetting( int *pn, struct tm* gt, FILE* AlmFile, double *Pos )
{
 long double jd;
 int gpsweek, timeweek;

 SetUserPos(Pos);

 jd = CalcJulianDay(gt);
 gpsweek = CalcGpsWeek( jd );

 timeweek = CalcToW(jd);

 ReadAlm(AlmFile);
 process_allsat(timeweek, gpsweek);
 SelectPrn(pn);

 return 0; // should return some more meaningful value?
}
//--------------------------------------------------------------------
/*++
 Thread related functions...
 --*/

void GPS_Func::CreateGpsThrd(void)
{
 gpsthrd = new GpsThrd(true,this);
 thrdtermed = false;
 gpsthrd->Resume();
}
void GPS_Func::TerminateThrd(void)
{
 if( gpsthrd )
  gpsthrd->Terminate();
}
// function to set the terminate status of sampling thread
void GPS_Func::SetThrdStatus(bool stat)
{
 thrdtermed = stat;
}
bool GPS_Func::NavThrdIsTermed()
{
 return (thrdtermed);
}




